from __future__ import annotations

from . import tables


def defragment(table: tables.AnyTable) -> tables.AnyTable:
    """Condense the underlying memory which backs the table to make
    it all contiguous. This makes many operations more efficient after
    defragmentation is complete.

    :param table: The table to defragment.
    :return: The defragmented table.
    """

    combined = table.table.combine_chunks()
    return table.__class__(table=combined)



from .__version__ import __version__
from .attributes import FloatAttribute, IntAttribute, StringAttribute
from .columns import (
    BinaryColumn,
    BooleanColumn,
    Byteslike,
    Column,
    Date32Column,
    Date64Column,
    Decimal128Column,
    Decimal256Column,
    DictionaryColumn,
    DurationColumn,
    FixedSizeBinaryColumn,
    FixedSizeListColumn,
    Float16Column,
    Float32Column,
    Float64Column,
    Int8Column,
    Int16Column,
    Int32Column,
    Int64Column,
    LargeBinaryColumn,
    LargeListColumn,
    LargeStringColumn,
    ListColumn,
    MapColumn,
    MetadataDict,
    MonthDayNanoIntervalColumn,
    NullColumn,
    RunEndEncodedColumn,
    StringColumn,
    StructColumn,
    SubTableColumn,
    Time32Column,
    Time64Column,
    TimestampColumn,
    UInt8Column,
    UInt16Column,
    UInt32Column,
    UInt64Column,
)
from .concat import concatenate
from .defragment import defragment
from .errors import (
    AttributeImmutableError,
    InvalidColumnDataError,
    InvalidColumnDefault,
    InvariantViolatedError,
    LinkageCombinationError,
    TableFragmentedError,
    TablesNotCompatibleError,
    ValidationError,
)
from .linkage import Linkage, MultiKeyLinkage, combine_linkages, combine_multilinkages
from .tables import (
    AnyTable,
    ArrowArrayProvider,
    AttributeValueType,
    DataSourceType,
    Table,
)
from .validators import Validator, and_, eq, ge, gt, is_in, le, lt

__all__ = [
    "AnyTable",
    "AttributeImmutableError",
    "AttributeValueType",
    "BinaryColumn",
    "BooleanColumn",
    "Byteslike",
    "Column",
    "DataSourceType",
    "Date32Column",
    "Date64Column",
    "Decimal128Column",
    "Decimal256Column",
    "DictionaryColumn",
    "DurationColumn",
    "FixedSizeBinaryColumn",
    "FixedSizeListColumn",
    "Float16Column",
    "Float32Column",
    "Float64Column",
    "FloatAttribute",
    "Int16Column",
    "Int32Column",
    "Int64Column",
    "Int8Column",
    "IntAttribute",
    "InvalidColumnDefault",
    "InvariantViolatedError",
    "LargeBinaryColumn",
    "LargeListColumn",
    "LargeStringColumn",
    "Linkage",
    "LinkageCombinationError",
    "ListColumn",
    "MapColumn",
    "MetadataDict",
    "MonthDayNanoIntervalColumn",
    "MultiKeyLinkage",
    "NullColumn",
    "RunEndEncodedColumn",
    "StringAttribute",
    "StringColumn",
    "StructColumn",
    "SubTableColumn",
    "Table",
    "TableFragmentedError",
    "TablesNotCompatibleError",
    "Time32Column",
    "Time64Column",
    "TimestampColumn",
    "UInt16Column",
    "UInt32Column",
    "UInt64Column",
    "UInt8Column",
    "ValidationError",
    "Validator",
    "__version__",
    "and_",
    "combine_linkages",
    "combine_multilinkages",
    "concatenate",
    "defragment",
    "eq",
    "ge",
    "gt",
    "is_in",
    "le",
    "lt",
    "InvalidColumnDataError",
    "ArrowArrayProvider",
]



from typing import Any

import pyarrow
import pyarrow.compute as pc

from . import errors


class Validator:
    """A Validator is a tool to validate that data in a
    :class:`pyarrow.Array` matches a predicate expression.

    :ivar func: The predicate function to use for validation. This
        must be a scalar or scalar aggregate function.
    :ivar args: The arguments to pass to the predicate function.
    :ivar label: A label to use when reporting validation errors.
    """

    def __init__(self, func: pc.Function, args: list[Any], label: str):
        """
        :param func: The predicate function to use for validation. This
            must be a scalar or scalar aggregate function.
        :param args: The arguments to pass to the predicate function.
        :param label: A label to use when reporting validation errors.
        """
        self._check_arity(func, args)
        if func.kind not in {"scalar_aggregate", "scalar"}:
            raise ValueError(f"Invalid function type {func.kind}, must be scalar or scalar_aggregate")
        self.func = func
        self.args = args
        self.label = label

    def _check_arity(self, func: pc.Function, args: list[Any]) -> None:
        if not func.arity == (len(args) + 1):
            raise ValueError("Invalid number of arguments")

    def evaluate(self, array: pyarrow.Array) -> pyarrow.Array:
        """
        Evaluates the predicate function on the given array.

        :param array: The array to evaluate.
        """
        return self.func.call([array, *self.args])

    def valid(self, array: pyarrow.Array) -> bool:
        """
        Returns True if the given array is valid, False otherwise.

        :param array: The array to validate.
        """
        if self.func.kind == "scalar_aggregate":
            return self.evaluate(array).as_py()  # type: ignore
        else:
            results = self.evaluate(array)
            if results.null_count == len(results):
                return True
            return pc.all(self.evaluate(array)).as_py()  # type: ignore

    def validate(self, array: pyarrow.Array) -> None:
        """
        Raises a :class:`ValidationError` if the given array is not valid.

        :param array: The array to validate.
        """
        if not self.valid(array):
            if self.func.kind == "scalar_aggregate":
                raise errors.ValidationError(f"array did not pass validator '{self.label}'")
            indices, failures = self.failures(array)
            if len(failures) == 1:
                index = indices[0].as_py()
                value = failures[0].as_py()
                msg = f"val={value}, index={index} failed validator '{self.label}'"
                raise errors.ValidationError(msg, failures)
            else:
                n_failed = len(failures)
                index = indices[0].as_py()
                value = failures[0].as_py()
                msg = (
                    f"validator '{self.label}' failed on {n_failed} values, "
                    + f"first failure: val={value}, index={index}"
                )
                raise errors.ValidationError(msg, failures)

    def failures(self, array: pyarrow.Array) -> tuple[pyarrow.Array, pyarrow.Array]:
        """
        Returns a tuple of two arrays, the first containing the indices of the invalid values,
        and the second containing the invalid values themselves.

        If the validator is a scalar aggregate function, raises a TypeError.
        """
        if self.func.kind == "scalar_aggregate":
            raise TypeError("Cannot get failures for scalar aggregate function")
        invalid = pc.invert(self.evaluate(array))
        indices = pc.indices_nonzero(invalid)
        invalid_values = pc.filter(array, invalid)
        return indices, invalid_values


class IsInValidator(Validator):
    """Executes the is_in validator, which is a special case because
    it takes the value set via a FunctionOption, rather than as a
    normal argument, so it has a different arity.

    """

    def __init__(self, args: list[Any], label: str):
        func = pc.get_function("is_in")
        super().__init__(func, args, label)

    def _check_arity(self, func: pc.Function, args: list[Any]) -> None:
        if not len(args) == 1:
            raise ValueError("Invalid number of arguments")

    def evaluate(self, array: pyarrow.Array) -> pyarrow.Array:
        return self.func.call([array], self.args[0])


class AndValidator(Validator):
    def __init__(self, validators: list[Validator], label: str):
        self.validators = validators
        func = pc.get_function("and")
        super().__init__(func, validators, label)

    def _check_arity(self, func: pc.Function, args: list[Any]) -> None:
        if not len(args) > 1:
            raise ValueError("Invalid number of arguments passed to and_")

    def evaluate(self, array: pyarrow.Array) -> pyarrow.Array:
        return pc.and_(*[v.evaluate(array) for v in self.validators])


def eq(val: Any) -> Validator:
    """
    Validator that all data in a column is equal to a given value.
    """
    func = pc.get_function("equal")
    label = f"eq({val})"
    return Validator(func, [val], label)


def lt(val: Any) -> Validator:
    """
    Validator that all data in a column is less than a given value.
    """
    func = pc.get_function("less")
    label = f"lt({val})"
    return Validator(func, [val], label)


def le(val: Any) -> Validator:
    """
    Validator that all data in a column is less than or equal to a given value.
    """
    func = pc.get_function("less_equal")
    label = f"le({val})"
    return Validator(func, [val], label)


def gt(val: Any) -> Validator:
    """
    Validator that all data in a column is greater than a given value.
    """
    func = pc.get_function("greater")
    label = f"gt({val})"
    return Validator(func, [val], label)


def ge(val: Any) -> Validator:
    """
    Validator that all data in a column is greater than or equal to a given value.
    """
    func = pc.get_function("greater_equal")
    label = f"ge({val})"
    return Validator(func, [val], label)


def is_in(val: Any, fail_on_null: bool = False) -> Validator:
    """Validator that all data in a column is in a given set.

    :param val: The set of values to check against. This can be a
        list, tuple, or pyarrow.Array. If it is a list or tuple, it
        will be converted to a pyarrow.Array.

    :param fail_on_null: If True, then nulls always trigger an
      error. Otherwise, they are matched to the value set, just like
      regular values.
    """
    label = f"is_in({val})"
    if not isinstance(val, pyarrow.Array):
        val = pyarrow.array(val)
    val = pc.SetLookupOptions(value_set=val, skip_nulls=fail_on_null)
    return IsInValidator([val], label)


def and_(*validators: Validator) -> Validator:
    """
    Validator that all data in a column passes all of the given validators.
    """
    label = "and({})".format(", ".join([v.label for v in validators]))
    return AndValidator(list(validators), label)



from typing import Iterator

import pyarrow as pa

from . import defragment, errors, tables


def concatenate(values: Iterator[tables.AnyTable], defrag: bool = True) -> tables.AnyTable:
    """Concatenate a collection of Tables into a single Table.

    All input Tables be of the same class, and have the same attribute
    values (if any).

    By default, results are compacted to be contiguous in memory,
    which involves a copy. In a tight loop, this can be very
    inefficient, so you can set the 'defrag' parameter to False to
    skip this compaction step, and instead call :func:`defragment` on
    the result after the loop is complete.

    :param values: An iterator of :class:`Table` instances to concatenate.
    :param defrag: Whether to compact the result to be contiguous in
        memory. Defaults to True.

    """
    batches = []
    first = True
    for v in values:
        batches += v.table.to_batches()
        if first:
            first_cls = v.__class__
            first_val = v
            first = False
        else:
            if v.__class__ != first_cls:
                raise errors.TablesNotCompatibleError("All tables must be the same class to concatenate")
            if not first_val._attr_equal(v):
                raise errors.TablesNotCompatibleError(
                    "All tables must have the same attribute values to concatenate"
                )

    if len(batches) == 0:
        raise ValueError("No values to concatenate")
    table = pa.Table.from_batches(batches)
    result = first_cls.from_pyarrow(table=table)
    if defrag:
        result = defragment.defragment(result)
    return result



from typing import Any, Optional

import pyarrow

from . import columns


class InvariantViolatedError(RuntimeError):
    """
    Exception raised when an invariant expectation is violated.
    """

    ...


class TableFragmentedError(InvariantViolatedError):
    """
    Exception raised when a Table is fragmented, and so an operation cannot
    be performed on it.
    """

    ...


class ValidationError(Exception):
    """
    Exception raised when data provided to a Table is invalid
    according to the Table's column validators.

    :ivar failures: An array of the invalid values that caused the
        exception to be raised.
    :vartype failures: pyarrow.Array
    """

    def __init__(self, message: str, failures: pyarrow.Array = None):
        super().__init__(message)
        self.failures = failures


class InvalidColumnDefault(Exception):
    """
    Exception raised when a column default is invalid.

    :ivar default_value: The invalid default value.
    :ivar dtype: The dtype of the column.
    :ivar column_name: The name of the column, if available. This may be None.
    """

    def __init__(self, default_value: Any, dtype: pyarrow.DataType, column_name: Optional[str] = None):
        msg = f"Invalid default value {repr(default_value)} for dtype {repr(dtype)}"
        if column_name is not None:
            msg += f" in column {repr(column_name)}"
        super().__init__(msg)
        self.default_value = default_value
        self.dtype = dtype
        self.column_name = column_name


class TablesNotCompatibleError(InvariantViolatedError):
    """Exception raised when two tables are not compatible for some
    operation like concatenation.
    """

    ...


class LinkageCombinationError(InvariantViolatedError):
    """Exception raised when linkages cannot be combined because they
    have incompatible table types or keys.
    """

    ...


class AttributeImmutableError(RuntimeError):
    """Exception raised when an attempt is made to modify an immutable
    attribute.
    """

    ...


class InvalidColumnDataError(Exception):
    """
    Exception raised when invalid data is provided to populate a particular column
    """

    def __init__(self, column: columns.Column, msg: str):
        self.column = column
        msg = f"Invalid data provided for column {column.name}: {msg}"
        super().__init__(msg)



from __future__ import annotations

import sys

if sys.version_info < (3, 10):
    from typing_extensions import TypeAlias
else:
    from typing import TypeAlias

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

import datetime
import decimal
from typing import Any, Callable, Dict, Generic, Optional, TypeVar, Union, overload

import pyarrow as pa

from . import errors, tables, validators

Byteslike: TypeAlias = Union[bytes, str]
MetadataDict: TypeAlias = Dict[Byteslike, Byteslike]


class Column:
    """A Column is an accessor for data in a Table, and also a
    descriptor for the Table's structure.

    This is a base class for all column types. It is not intended to
    be used directly; instead, use one of its subclasses.

    Columns implement the descriptor protocol, so they should be used
    as class attributes on a Table subclass.

    :param dtype: The pyarrow data type of the column.
    :param nullable: Whether the column can contain null values.
    :param metadata: A dictionary of metadata to attach to the column.
    :param validator: A validator to use when setting the column.
    :param default: A default value to use when setting the column.

    :ivar dtype: The pyarrow data type of the column.
    :ivar nullable: Whether the column can contain null values.
    :ivar metadata: A dictionary of metadata to attach to the column.
    :ivar validator: A validator to use when setting the column.
    :ivar default: A default value to use when setting the column.
    """

    def __init__(
        self,
        dtype: pa.DataType,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
        default: Union[None, Callable[[], Any], Any] = None,
    ):
        self.dtype = dtype
        self.nullable = nullable
        self.metadata = metadata
        self.validator = validator

        self.default = default
        if self.default is not None:
            self._default_is_callable = callable(self.default)
            if self._default_is_callable:
                self._default_fn = self.default
            else:
                try:
                    self._default_scalar = pa.scalar(self.default, self.dtype)
                except (pa.ArrowInvalid, pa.ArrowTypeError, OverflowError) as e:
                    raise errors.InvalidColumnDefault(self.default, self.dtype) from e

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.Array:
        ...

    def __get__(self, obj: Union[tables.Table, None], objtype: type) -> Union[Self, pa.Array]:
        """

        Gets the Column object from a Table class, or the associated data from a Table instance.

        This method is part of the `descriptor protocol <https://docs.python.org/3/howto/descriptor.html>`_.

        """
        if obj is None:
            return self
        return obj.table.column(self.name)

    def __set__(self, obj: tables.Table, value: pa.Array) -> None:
        """
        Sets the data for this column on a Table instance.

        This method is part of the `descriptor protocol <https://docs.python.org/3/howto/descriptor.html>`_.
        """
        idx = obj.table.schema.get_field_index(self.name)
        value = self.fill_default(value)
        obj.table = obj.table.set_column(idx, self.pyarrow_field(), [value])

    def __set_name__(self, owner: type, name: str) -> None:
        """
        Sets the name of the column.

        This method is part of the `descriptor protocol <https://docs.python.org/3/howto/descriptor.html>`_.
        """
        self.name = name

    def fill_default(self, array: pa.Array) -> pa.Array:
        """
        Fills null values in the array with the Column's default value.
        """
        if self.default is None or array.null_count == 0:
            return array
        if self._default_is_callable:
            null_count = array.null_count
            try:
                fill_values = pa.array([self._default_fn() for _ in range(null_count)], self.dtype)
            except (pa.ArrowInvalid, pa.ArrowTypeError, OverflowError) as e:
                raise errors.InvalidColumnDefault(self._default_fn, self.dtype, self.name) from e
            null_mask = pa.compute.is_null(array)
            return pa.compute.replace_with_mask(array, null_mask, fill_values)
        else:
            return array.fill_null(self._default_scalar)

    def pyarrow_field(self) -> pa.Field:
        """
        Returns a pyarrow Field object for this column.
        """
        return pa.field(self.name, self.dtype, self.nullable, self.metadata)

    def _load(
        self,
        data: Union[tables.DataSourceType, None],
        size_hint: Optional[int],
    ) -> Optional[pa.Array]:
        """Load a data source as a pyarrow Array of data which
        matches the column's schema. Defaults are not poulated.

        size_hint is an optional expected size of the input data. If provided:
         - For input of None, an array of nulls is returned.
         - For all other inputs, the input's length is checked.

        If the column is nullable and the input value is None, then
        this returns None.  Otherwise, it will either return a pyarrow
        Array, or raise an error.
        """
        if data is None:
            if not self.nullable and self.default is None:
                raise errors.InvalidColumnDataError(
                    self, "received no data, but field is not nullable and has no default"
                )

            # Either it's nullable, or has a default; either way, we can proceed.
            if size_hint is None:
                # We can't even make a null array of the appropriate
                # size, so we're forced to return None.
                return None

            return self._nulls(size_hint)

        return pa.array(data, type=self.dtype)

    def _nulls(self, n: int) -> pa.Array:
        """Return an array of nulls of the appropriate size."""
        return pa.nulls(n, type=self.dtype)


T = TypeVar("T", bound=tables.Table)


class SubTableColumn(Column, Generic[T]):
    """
    A column which represents an embedded quivr table.

    :param table_type: The type of the table to embed.
    :param nullable: Whether the column can contain null values.
    :param metadata: A dictionary of metadata to attach to the column.
    """

    def __init__(self, table_type: type[T], nullable: bool = False, metadata: Optional[MetadataDict] = None):
        self.table_type = table_type

        self.schema = table_type.schema
        if nullable:
            # permit nulls in fields. This is necessary for the
            # case where the field is not-nullable, but the table
            # is nullable.
            fields = []
            for field in self.schema:
                if field.nullable:
                    fields.append(field)
                else:
                    fields.append(field.with_nullable(True))
            self.schema = pa.schema(fields)

        dtype = pa.struct(self.schema)
        super().__init__(dtype, nullable=nullable, metadata=metadata)

    def __set__(self, obj: tables.Table, value: T) -> None:
        # Propagate the value's metadata through to the parent
        metadata = obj.table.schema.metadata
        value_meta = value.table.schema.metadata
        if value_meta is not None:
            for key, val in value_meta.items():
                key = (self.name + "." + key.decode("utf-8")).encode("utf-8")
                metadata[key] = val

        idx = obj.table.schema.get_field_index(self.name)
        obj.table = obj.table.replace_schema_metadata(metadata)
        data = value.to_structarray()
        if self.nullable:
            # rewrite the structarray's type to make all fields
            # nullable. This is necessary for the case where the
            # field is not-nullable, but the table is nullable.
            fields = []
            for field in data.type:
                fields.append(field.with_nullable(True))
            data = pa.StructArray.from_arrays(data.flatten(), fields=fields)
        obj.table = obj.table.set_column(idx, self.pyarrow_field(), [data])

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> T:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, T]:
        if obj is None:
            return self
        array = obj.table.column(self.name)

        metadata = self.metadata
        if metadata is None:
            metadata = {}
        metadata.update(obj._metadata_for_column(self.name))  # type: ignore

        schema = self.schema.with_metadata(metadata)

        subtable = pa.Table.from_arrays(array.flatten(), schema=schema)
        return self.table_type.from_pyarrow(subtable, permit_nulls=self.nullable)

    def _nulls(self, n: int) -> pa.Array:
        """Return an array of nulls of the appropriate size."""
        # Quite unfortunate: pyarrow doesn't support creating a struct
        # array of nulls; it incorrectly populates all the fields with
        # nulls (including non-nullable ones!) so we need to do this
        # manually.
        #
        # See: https://github.com/apache/arrow/issues/37072
        return pa.array([None] * n, type=self.dtype)


class Int8Column(Column):
    """
    A column for storing 8-bit integers.
    """

    primitive_dtype = pa.int8()

    def __init__(
        self,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
        default: Union[None, Callable[[], int], int] = None,
    ):
        super().__init__(
            pa.int8(), nullable=nullable, metadata=metadata, validator=validator, default=default
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.Int8Array:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.Int8Array]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class Int16Column(Column):
    """
    A column for storing 16-bit integers.
    """

    primitive_dtype = pa.int16()

    def __init__(
        self,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
        default: Union[None, Callable[[], int], int] = None,
    ):
        super().__init__(
            pa.int16(), nullable=nullable, metadata=metadata, validator=validator, default=default
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.Int16Array:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.Int16Array]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class Int32Column(Column):
    """
    A column for storing 32-bit integers.
    """

    primitive_dtype = pa.int32()

    def __init__(
        self,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
        default: Union[None, Callable[[], int], int] = None,
    ):
        super().__init__(
            pa.int32(), nullable=nullable, metadata=metadata, validator=validator, default=default
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.Int32Array:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.Int32Array]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class Int64Column(Column):
    """
    A column for storing 64-bit integers.
    """

    primitive_dtype = pa.int64()

    def __init__(
        self,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
        default: Union[None, Callable[[], int], int] = None,
    ):
        super().__init__(
            pa.int64(), nullable=nullable, metadata=metadata, validator=validator, default=default
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.Int64Array:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.Int64Array]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class UInt8Column(Column):
    """
    A column for storing 8-bit unsigned integers.
    """

    primitive_dtype = pa.uint8()

    def __init__(
        self,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
        default: Union[None, Callable[[], int], int] = None,
    ):
        super().__init__(
            pa.uint8(), nullable=nullable, metadata=metadata, validator=validator, default=default
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.UInt8Array:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.UInt8Array]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class UInt16Column(Column):
    """
    A column for storing 16-bit unsigned integers.
    """

    primitive_dtype = pa.uint16()

    def __init__(
        self,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
        default: Union[None, Callable[[], int], int] = None,
    ):
        super().__init__(
            pa.uint16(), nullable=nullable, metadata=metadata, validator=validator, default=default
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.UInt16Array:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.UInt16Array]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class UInt32Column(Column):
    """
    A column for storing 32-bit unsigned integers.
    """

    primitive_dtype = pa.uint32()

    def __init__(
        self,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
        default: Union[None, Callable[[], int], int] = None,
    ):
        super().__init__(
            pa.uint32(), nullable=nullable, metadata=metadata, validator=validator, default=default
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.UInt32Array:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.UInt32Array]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class UInt64Column(Column):
    """
    A column for storing 64-bit unsigned integers.
    """

    primitive_dtype = pa.uint64()

    def __init__(
        self,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
        default: Union[None, Callable[[], int], int] = None,
    ):
        super().__init__(
            pa.uint64(), nullable=nullable, metadata=metadata, validator=validator, default=default
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.UInt64Array:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.UInt64Array]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class Float16Column(Column):
    """A column for storing 16-bit floating point numbers.

    16-bit floating point arrays have limited support in Arrow, and
    thus limited support in quivr.

    In particular:
    - They cannot be written to or read from Parquet files.

    - They cannot be constructed from Python floats in a natural way
      (one must use numpy.float16).

    - They don't support quivr column default values (because they are
      not supported by Arrow's compute functions).
    """

    primitive_dtype = pa.float16()

    def __init__(
        self,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
    ):
        super().__init__(pa.float16(), nullable=nullable, metadata=metadata, validator=validator)

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.lib.HalfFloatArray:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.lib.HalfFloatArray]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class Float32Column(Column):
    """
    A column for storing 32-bit floating point numbers.
    """

    primitive_dtype = pa.float32()

    def __init__(
        self,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
        default: Union[None, Callable[[], float], float] = None,
    ):
        super().__init__(
            pa.float32(), nullable=nullable, metadata=metadata, validator=validator, default=default
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.lib.FloatArray:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.lib.FloatArray]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class Float64Column(Column):
    """
    A column for storing 64-bit floating point numbers.
    """

    primitive_dtype = pa.float64()

    def __init__(
        self,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
        default: Union[None, Callable[[], float], float] = None,
    ):
        super().__init__(
            pa.float64(), nullable=nullable, metadata=metadata, validator=validator, default=default
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.lib.DoubleArray:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.lib.DoubleArray]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class BooleanColumn(Column):
    """A column for storing booleans."""

    primitive_dtype = pa.bool_()

    def __init__(
        self,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
        default: Union[None, Callable[[], bool], bool] = None,
    ):
        super().__init__(
            pa.bool_(), nullable=nullable, metadata=metadata, validator=validator, default=default
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.BooleanArray:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[pa.BooleanArray, Self]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class StringColumn(Column):
    """A column for storing strings.

    This can be used to store strings of any length, but it is not
    recommended for storing very long strings (over 2GB, for
    example). For long strings, use LargeStringColumn instead.

    """

    primitive_dtype = pa.string()

    def __init__(
        self,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
        default: Union[None, str, Callable[[], str]] = None,
    ):
        super().__init__(
            pa.string(), nullable=nullable, metadata=metadata, validator=validator, default=default
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.StringArray:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[pa.StringArray, Self]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class LargeBinaryColumn(Column):
    r"""
    A column for storing large binary objects (over 2\ :sup:`31` bytes long). Large binary data is stored in
    variable-length chunks.
    """

    def __init__(
        self,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
        default: Union[None, bytes, Callable[[], bytes]] = None,
    ):
        super().__init__(
            pa.large_binary(), nullable=nullable, metadata=metadata, validator=validator, default=default
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.LargeBinaryArray:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.LargeBinaryArray]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class LargeStringColumn(Column):
    r"""
    A column for storing large strings (over 2\ :sup:`31` bytes long). Large string data is stored in
    variable-length chunks.
    """

    def __init__(
        self,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
        default: Union[None, str, Callable[[], str]] = None,
    ):
        super().__init__(
            pa.large_string(), nullable=nullable, metadata=metadata, validator=validator, default=default
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.LargeStringArray:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.LargeStringArray]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class Date32Column(Column):
    """A column for storing dates.

    Internally, this column stores dates as 32-bit integers which
    represent time since the UNIX epoch.

    """

    def __init__(
        self,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
    ):
        super().__init__(pa.date32(), nullable=nullable, metadata=metadata, validator=validator)

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.Date32Array:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.Date32Array]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class Date64Column(Column):
    """A column for storing dates.

    Internally, this column stores dates as 64-bit integers which
    represent time since the UNIX epoch in milliseconds, where the
    values are evenly divisible by 86,400,000.
    """

    def __init__(
        self,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
    ):
        super().__init__(pa.date64(), nullable=nullable, metadata=metadata, validator=validator)

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.Date64Array:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.Date64Array]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class TimestampColumn(Column):
    """A column for storing timestamps.

    Timestamp data can be stored in one of four units:
      - seconds
      - milliseconds
      - microseconds
      - nanoseconds

    Internally, a timestamp is a 64-bit integer which represents time
    since 1970-01-01 00:00:00 UTC (the UNIX epoch), and is optionally
    annotated with a timezone.

    Timestamp values do not include any leap seconds (in other words,
    all days are considered 86,400 seconds long).

    :param unit: The unit of time to use for the timestamp. Valid
        values are "s", "ms", "us", and "ns".
    :param tz: An optional timezone to associate with the timestamp.
    :param nullable: Whether the column can contain null values.
    :param metadata: Optional metadata to associate with the column.
    :param validator: An optional validator to apply to the column.
    :param default: An optional default value for the column. This
        can be a scalar value or a callable that takes no arguments.
    """

    def __init__(
        self,
        unit: str,
        tz: Optional[str] = None,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
        default: Union[None, datetime.datetime, Callable[[], datetime.datetime]] = None,
    ):
        super().__init__(
            pa.timestamp(unit, tz), nullable=nullable, metadata=metadata, validator=validator, default=default
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.TimestampArray:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.TimestampArray]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class Time32Column(Column):
    """
    A column for storing time values.

    Time data can be stored in one of two units for this 32-bit type:
        - seconds
        - milliseconds

    Internally, a time32 value is a 32-bit integer which an elapsed time since midnight.

    :param unit: The unit of time to use for the time. Valid values are "s" and "ms".
    :param nullable: Whether the column can contain null values.
    :param metadata: Optional metadata to associate with the column.
    :param validator: An optional validator to apply to the column.
    :param default: An optional default value for the column. This
        can be a scalar value or a callable that takes no arguments.
    """

    def __init__(
        self,
        unit: str,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
        default: Union[None, datetime.time, Callable[[], datetime.time]] = None,
    ):
        super().__init__(
            pa.time32(unit), nullable=nullable, metadata=metadata, validator=validator, default=default
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.Time32Array:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.Time32Array]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class Time64Column(Column):
    """
    A column for storing time values with high precision.

    Time data can be stored in one of two units for this 64-bit type:
        - microseconds
        - nanoseconds

    Internally, a time64 value is a 64-bit integer which an elapsed time since midnight.

    :param unit: The unit of time to use for the time. Valid values are "us" and "ns".
    :param nullable: Whether the column can contain null values.
    :param metadata: Optional metadata to associate with the column.
    :param validator: An optional validator to apply to the column.
    :param default: An optional default value for the column. This
        can be a scalar value or a callable that takes no arguments.
    """

    def __init__(
        self,
        unit: str,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
        default: Union[None, datetime.time, Callable[[], datetime.time]] = None,
    ):
        super().__init__(
            pa.time64(unit), nullable=nullable, metadata=metadata, validator=validator, default=default
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.Time64Array:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.Time64Array]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class DurationColumn(Column):
    """
    A column which stores an absolute length of time.

    Internally, a duration value is always represented as an 64-bit integer.
    :param unit: The unit of time to use for the duration. Valid values are "s", "ms", "us", and "ns".
    :param nullable: Whether the column can contain null values.
    :param metadata: Optional metadata to associate with the column.
    :param validator: An optional validator to apply to the column.
    """

    def __init__(
        self,
        unit: str,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
    ):
        super().__init__(pa.duration(unit), nullable=nullable, metadata=metadata, validator=validator)

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.DurationArray:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.DurationArray]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class MonthDayNanoIntervalColumn(Column):
    """A column for storing calendar intervals (an elapsed number of months,
    days, and nanoseconds).

    Internally, a month_day_nano_interval value is a 96-bit
    integer. 32 bits are devoted to months and days, and 64 bits are
    devoted to nanoseconds.

    Leap seconds are ignored.

    """

    def __init__(
        self,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
    ):
        super().__init__(
            pa.month_day_nano_interval(), nullable=nullable, metadata=metadata, validator=validator
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.MonthDayNanoIntervalArray:
        ...

    def __get__(
        self, obj: Optional[tables.Table], objtype: type
    ) -> Union[Self, pa.MonthDayNanoIntervalArray]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class BinaryColumn(Column):
    """A column for storing opaque binary data."""

    def __init__(
        self,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
        default: Union[None, bytes, Callable[[], bytes]] = None,
    ):
        super().__init__(
            pa.binary(-1), nullable=nullable, metadata=metadata, validator=validator, default=default
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.BinaryArray:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.BinaryArray]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class FixedSizeBinaryColumn(Column):
    """A column for storing opaque fixed-size binary data.

    :param byte_width: The number of bytes per value.
    :param nullable: Whether the column can contain null values.
    :param metadata: Optional metadata to associate with the column.
    :param validator: An optional validator to apply to the column.
    :param default: An optional default value for the column. This
        can be a scalar value or a callable that takes no arguments.
    """

    def __init__(
        self,
        byte_width: int,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
        default: Union[None, bytes, Callable[[], bytes]] = None,
    ):
        self.byte_width = byte_width
        super().__init__(
            pa.binary(byte_width), nullable=nullable, metadata=metadata, validator=validator, default=default
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.FixedSizeBinaryArray:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.FixedSizeBinaryArray]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class Decimal128Column(Column):
    """A column for storing arbitrary-precision decimal numbers.

    Arrow decimals are fixed-point decimal numbers encoded as a scaled
    integer. The precision is the number of significant digits that the
    decimal type can represent; the scale is the number of digits after
    the decimal point (note the scale can be negative).

    As an example, ``Decimal128Column(7, 3)`` can exactly represent the numbers
    1234.567 and -1234.567 (encoded internally as the 128-bit integers
    1234567 and -1234567, respectively), but neither 12345.67 nor
    123.4567.

    ``DecimalColumn(5, -3)`` can exactly represent the number 12345000
    (encoded internally as the 128-bit integer 12345), but neither
    123450000 nor 1234500.

    If you need a precision higher than 38 significant digits,
    consider using :class:`Decimal256Column`.

    :param precision: The number of significant digits.
    :param scale: The number of digits after the decimal point.
    :param nullable: Whether the column can contain nulls.
    :param metadata: A dictionary of metadata to attach to the column.
    :param default: An optional default value for the column. This
        can be a scalar value or a callable that takes no arguments.
    """

    def __init__(
        self,
        precision: int,
        scale: int,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        default: Union[None, decimal.Decimal, Callable[[], decimal.Decimal]] = None,
    ):
        super().__init__(
            pa.decimal128(precision, scale), nullable=nullable, metadata=metadata, default=default
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.Decimal128Array:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.Decimal128Array]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class Decimal256Column(Column):
    """A column for storing arbitrary-precision decimal numbers.

    Arrow decimals are fixed-point decimal numbers encoded as a scaled
    integer. The precision is the number of significant digits that the
    decimal type can represent; the scale is the number of digits after
    the decimal point (note the scale can be negative).

    Values are stored as 256-bit integers.

    :param precision: The number of significant digits.
    :param scale: The number of digits after the decimal point.
    :param nullable: Whether the column can contain nulls.
    :param metadata: A dictionary of metadata to attach to the column.
    :param default: An optional default value for the column. This
        can be a scalar value or a callable that takes no arguments.
    """

    def __init__(
        self,
        precision: int,
        scale: int,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        default: Union[None, decimal.Decimal, Callable[[], decimal.Decimal]] = None,
    ):
        super().__init__(
            pa.decimal256(precision, scale), nullable=nullable, metadata=metadata, default=default
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.Decimal256Array:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.Decimal256Array]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class NullColumn(Column):
    """A column for storing null values.

    Nulls are represented as a single bit, and do not take up any
    memory space.

    """

    def __init__(
        self,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
    ):
        super().__init__(pa.null(), nullable=nullable, metadata=metadata, validator=validator)

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.NullArray:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.NullArray]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


# Complex types follow


class ListColumn(Column):
    """A column for storing variably-sized lists of values.

    The values in the list can be of any type.

    Note that all quivr Tables are storing lists of values, so this
    column type is only useful for storing lists of lists.

    :param value_type: The type of the values in the list.
    :param nullable: Whether the list can contain null values.
    :param metadata: A dictionary of metadata to attach to the column.
    :param validator: A validator to run against the column's values.

    """

    def __init__(
        self,
        value_type: Union[pa.DataType, pa.Field, Column],
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
    ):
        if isinstance(value_type, Column):
            value_type = value_type.dtype
        super().__init__(pa.list_(value_type, -1), nullable=nullable, metadata=metadata, validator=validator)

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.ListArray:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.ListArray]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class FixedSizeListColumn(Column):
    """A column for storing lists of values of a fixed size.

    The values in the list can be of any type.

    Note that all quivr Tables are storing lists of values, so this
    column type is only useful for storing lists of lists.

    :param value_type: The type of the values in the list.
    :param list_size: The size of the list. Must be greater than 0.
    :param nullable: Whether the list can contain null values.
    :param metadata: A dictionary of metadata to attach to the column.
    :param validator: A validator to run against the column's values.
    """

    def __init__(
        self,
        value_type: Union[pa.DataType, pa.Field, Column],
        list_size: int,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
    ):
        if list_size <= 0:
            raise ValueError("list_size must be greater than 0")
        if isinstance(value_type, Column):
            value_type = value_type.dtype
        super().__init__(
            pa.list_(value_type, list_size), nullable=nullable, metadata=metadata, validator=validator
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.FixedSizeListArray:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.FixedSizeListArray]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class LargeListColumn(Column):
    r"""A column for storing large lists of values (over 2\ :sup:`31` objects).

    Unless you need to represent data with more than 2**31 elements,
    prefer ListColumn.

    The values in the list can be of any type.

    Note that all quivr Tables are storing lists of values, so this
    column type is only useful for storing lists of lists.

    :param value_type: The type of the values in the list.
    :param nullable: Whether the list can contain null values.
    :param metadata: A dictionary of metadata to attach to the column.
    :param validator: A validator to run against the column's values.
    """

    def __init__(
        self,
        value_type: Union[pa.DataType, pa.Field, Column],
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
    ):
        if isinstance(value_type, Column):
            value_type = value_type.dtype
        super().__init__(pa.large_list(value_type), nullable=nullable, metadata=metadata, validator=validator)

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.LargeListArray:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.LargeListArray]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class MapColumn(Column):
    """A column for storing maps of key-value pairs.

    The keys and values can be of any type, as long as the keys are
    hashable and unique.

    :param key_type: The type of the keys in the map.
    :param item_type: The type of the values in the map.
    :param nullable: Whether the map can contain null key-value pairs.
    :param metadata: A dictionary of metadata to attach to the column.
    :param validator: A validator to run against the column's values.
    """

    def __init__(
        self,
        key_type: Union[pa.DataType, pa.Field, Column],
        item_type: Union[pa.DataType, pa.Field, Column],
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
    ):
        if isinstance(key_type, Column):
            key_type = key_type.dtype
        if isinstance(item_type, Column):
            item_type = item_type.dtype
        super().__init__(
            pa.map_(key_type, item_type), nullable=nullable, metadata=metadata, validator=validator
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.MapArray:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.MapArray]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class DictionaryColumn(Column):
    """A column for storing dictionary-encoded values.

    This is intended for use with categorical data. See MapColumn for a
    more general mapping type.

    :param index_type: The type of the dictionary indices. Must be an integer type.
    :param value_type: The type of the values in the dictionary.
    :param ordered: Whether the dictionary is ordered.
    :param nullable: Whether the dictionary can contain null values.
    :param metadata: A dictionary of metadata to attach to the column.
    :param validator: A validator to run against the column's values.
    """

    def __init__(
        self,
        index_type: pa.DataType,
        value_type: Union[pa.DataType, pa.Field, Column],
        ordered: bool = False,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
    ):
        if isinstance(index_type, Column):
            index_type = index_type.dtype
        if isinstance(value_type, Column):
            value_type = value_type.dtype
        super().__init__(
            pa.dictionary(index_type, value_type, ordered=ordered),
            nullable=nullable,
            metadata=metadata,
            validator=validator,
        )

    @overload
    def __get__(self, obj: None, objtype: type) -> Self:
        ...

    @overload
    def __get__(self, obj: tables.Table, objtype: type) -> pa.DictionaryArray:
        ...

    def __get__(self, obj: Optional[tables.Table], objtype: type) -> Union[Self, pa.DictionaryArray]:
        if obj is None:
            return self
        return obj.table[self.name].combine_chunks()


class StructColumn(Column):
    """A column for storing structured data.

    In general, prefer to define Tables and use their as_column method
    instead of using StructColumn.

    :param fields: The fields in the struct.
    :param nullable: Whether the struct can contain null values.
    :param metadata: A dictionary of metadata to attach to the column.
    :param validator: A validator to run against the column's values.

    """

    def __init__(
        self,
        fields: list[pa.Field],
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
    ):
        super().__init__(pa.struct(fields), nullable=nullable, metadata=metadata, validator=validator)


class RunEndEncodedColumn(Column):
    """A column for storing run-end encoded data.

    This is a special column type that is used to efficiently store
    highly ordered data. Internally, the data is stored as two
    buffers:

    - An array of values, with all consecutive runs of the same value
      reduced to a single element
    - An array of run lengths, with the length of each run

    This is more compact than storing the redundant values and also
    allow for very efficient computations like aggregations upon the
    data.


    :param run_end_type: The type of the run-end encoded values. Must be a 16-, 32-, or 64-bit integer type.
    :param value_type: The type of the values in the run-end encoded data.
    :param nullable: Whether the data can contain null values.
    :param metadata: A dictionary of metadata to attach to the column.
    :param validator: A validator to run against the column's values.
    """

    def __init__(
        self,
        run_end_type: pa.DataType,
        value_type: pa.DataType,
        nullable: bool = False,
        metadata: Optional[MetadataDict] = None,
        validator: Optional[validators.Validator] = None,
    ):
        super().__init__(
            pa.run_end_encoded(run_end_type, value_type),
            nullable=nullable,
            metadata=metadata,
            validator=validator,
        )



from __future__ import annotations

import os
import sys
import warnings
from io import IOBase

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

from typing import (
    Any,
    ClassVar,
    Iterator,
    List,
    Optional,
    Protocol,
    Type,
    TypeAlias,
    TypeVar,
    Union,
    cast,
)

import numpy as np
import numpy.typing as npt
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.csv
import pyarrow.feather
import pyarrow.parquet

from . import attributes, columns, errors, schemagraph, validators


class ArrowArrayProvider(Protocol):
    """
    A Protocol which describes objects that support the Arrow custom array extension protocol.
    """

    def __arrow_array__(self, type: Optional[pa.DataType] = None) -> pa.Array:
        ...


AttributeValueType: TypeAlias = Union[int, float, str]
DataSourceType: TypeAlias = Union[
    pa.Array, list[Any], "Table", pd.Series, npt.NDArray[Any], ArrowArrayProvider
]
AnyTable = TypeVar("AnyTable", bound="Table")

# If a table uses any of the following names, it will break quivr
# internals entirely, so they must be rejected.
_FORBIDDEN_COLUMN_NAMES = {
    "table",
    "schema",
    "_quivr_subtables",
    "_quivr_attributes",
    "_column_validators",
}


class Table:
    """Table is the primary data structure in quivr.

    Tables are used to represent tabular data with a fixed schema. The
    schema is defined by subclassing Table, providing Column objects
    as class attributes. The Table class will then generate a pyarrow
    schema from those columns.

    Table instances can then be created from data, either by passing
    in a pyarrow Table, or by passing in data in a variety of other
    formats. The data will be validated against the schema, and
    converted to a pyarrow Table.

    Table instances are immutable, but can be sliced, filtered,
    sorted, or otherwise manipulated, resulting in new Table
    instances.

    :cvar schema: The pyarrow schema for this table.
    :vartype schema: pyarrow.Schema

    :ivar table: The underlying :class:`pyarrow.Table` for this Table instance.
    :vartype table: pyarrow.Table
    """

    schema: ClassVar[pa.Schema]

    table: pa.Table

    _quivr_subtables: ClassVar[dict[str, columns.SubTableColumn[Any]]]
    _quivr_attributes: ClassVar[dict[str, attributes.Attribute[Any]]]
    _column_validators: ClassVar[dict[str, validators.Validator]]

    def __init_subclass__(cls: Type["Table"], **kwargs: Any):
        fields = []
        column_validators = {}
        subtables = {}
        attrs = {}
        for name, obj in cls.__dict__.items():
            if name in _FORBIDDEN_COLUMN_NAMES:
                raise AttributeError(
                    f"Invalid column name {name} in {cls.__name__}: {name} is a reserved name"
                )
            if isinstance(obj, columns.Column):
                fields.append(obj.pyarrow_field())
                if obj.validator is not None:
                    column_validators[name] = obj.validator
                if isinstance(obj, columns.SubTableColumn):
                    subtables[name] = obj
            elif isinstance(obj, attributes.Attribute):
                attrs[name] = obj

        # Generate a pyarrow schema
        schema = pa.schema(fields)
        cls.schema = schema

        # Keep track of subtables
        cls._quivr_subtables = subtables

        # Add attributes
        cls._quivr_attributes = attrs

        # Add validators
        cls._column_validators = column_validators

        super().__init_subclass__(**kwargs)

    def __init__(self, table: pa.Table, **kwargs: AttributeValueType):
        self.table = table
        for name, value in kwargs.items():
            if name in self._quivr_attributes:
                setattr(self, name, value)
            else:
                raise AttributeError(f"{self.__class__.__name__} has no attribute {name}")
        for name in self._quivr_attributes:
            # Ensure all attributes are set or have a default
            getattr(self, name)

    @classmethod
    def from_pyarrow(
        cls,
        table: pa.Table,
        validate: bool = True,
        permit_nulls: bool = False,
        **kwargs: AttributeValueType,
    ) -> Self:
        """Create a new table from a pyarrow Table.

        This is a convenience method which can be used to create a
        Table from a pyarrow Table. It can also accept keyword-style
        arguments to set attributes on the table.

        When serializing to pyarrow, the table's schema metadata
        encodes the value of attributes. This method will use that
        metadata to set the attributes on the table if it is
        available. If any attributes are provided as keyword
        arguments, they override any values in the metadata.

        :param table: The pyarrow Table to create the table from.
        :param validate: Whether to validate the table against the
            schema, and run any column validators.
        :param permit_nulls: Whether to permit null values in the
            table. If True, nulls will be permitted, even in non-nullable
            fields. This is used when a Table is used as a nullable subtable.
        :param \\**kwargs: Keyword arguments to set attributes on the
            table.
        :type \\**kwargs: :obj:`AttributeValueType`
        :return: A new Table instance.
        """
        schema = cls.schema
        if permit_nulls:
            # Rewrite the schema to permit nulls
            fields = []
            for field in schema:
                if field.nullable:
                    fields.append(field)
                else:
                    fields.append(field.with_nullable(True))
            schema = pa.schema(fields, metadata=schema.metadata)

        # Absorb metadata from the table
        schema = schema.with_metadata(table.schema.metadata)

        table = table.cast(schema)
        instance = cls(table, **kwargs)
        if validate:
            instance.validate()
        return instance

    @classmethod
    def from_data(
        cls,
        data: Union[pa.Table, dict[str, Any], list[Any], pd.DataFrame, None] = None,
        validate: bool = True,
        **kwargs: Union[AttributeValueType, DataSourceType],
    ) -> Self:
        """Create an instance of the Table and populate it with data.

        This is a convenience method which tries to infer the right
        underlying constructors to use based on the type of data. It
        can also accept keyword-style arguments to pass data in. If
        you know the data's structure well in advance, the more
        precise constructors (from_arrays, from_pylist, etc) should be
        preferred.

        If the validate argument is True, the data will be validated
        against the table's schema. If validation fails, a
        :class:`ValidationError` will be raised.

        """
        warnings.warn(DeprecationWarning("Table.from_data will be removed in quivr version 0.7"))
        if data is None:
            instance = cls.from_kwargs(validate=validate, **kwargs)
        else:
            attrib_kwargs = cls._attribute_kwargs_from_kwargs(kwargs)
            if isinstance(data, pa.Table):
                instance = cls.from_pyarrow(data, False, **attrib_kwargs)
            elif isinstance(data, dict):
                instance = cls.from_pydict(data, **attrib_kwargs)
            elif isinstance(data, list):
                if len(data) == 0:
                    instance = cls.from_rows(data, **attrib_kwargs)
                elif isinstance(data[0], dict):
                    instance = cls.from_rows(data, **attrib_kwargs)
                elif isinstance(data[0], list):
                    instance = cls.from_lists(data, **attrib_kwargs)
                else:
                    raise TypeError(f"Unsupported type: {type(data[0])}")
            elif isinstance(data, pd.DataFrame):
                instance = cls.from_dataframe(data, validate, **attrib_kwargs)
            else:
                raise TypeError(f"Unsupported type: {type(data)}")

        if validate:
            instance.validate()
        return instance

    @classmethod
    def _attribute_kwargs_from_kwargs(
        cls, kwargs: dict[str, Union[AttributeValueType, DataSourceType]]
    ) -> dict[str, AttributeValueType]:
        attrib_kwargs: dict[str, AttributeValueType] = {}
        for k, v in kwargs.items():
            if k not in cls._quivr_attributes:
                raise TypeError(f"Unexpected keyword argument: {k}")
            if not isinstance(v, cls._quivr_attributes[k]._type):
                raise TypeError(f"Expected {cls._quivr_attributes[k]._type}, got {type(v)} for {k}")
            attrib_kwargs[k] = v
        return attrib_kwargs

    @classmethod
    def as_column(
        cls, nullable: bool = True, metadata: Optional[columns.MetadataDict] = None
    ) -> columns.SubTableColumn[Self]:
        """Embed the Table as a column in another Table.

        This method is the primary way to achieve composition of Tables with quivr.

        :param nullable: Whether the column can contain nulls. Note
            that this refers to whether an entire row - all columns -
            can be null, not whether a single value in one column of
            this table can be null. That is controlled entirely by
            this Table class's columns.

        :param metadata: Metadata to attach to the column.

        """
        return columns.SubTableColumn(cls, nullable=nullable, metadata=metadata)

    @classmethod
    def from_kwargs(cls, validate: bool = True, **kwargs: Union[DataSourceType, AttributeValueType]) -> Self:
        """Create a Table instance from keyword arguments.

        Each keyword argument corresponds to a column in the Table.

        The keys should correspond to column names or attribute names.

        For columns, the values should be arrays, lists, or pyarrow Arrays.

        For attributes, the values should be the appropriate type for that attribute.

        :param validate: If (the default), run column validators on all input data.
        :param \\**kwargs: The data to populate the table with.
        :type \\**kwargs: Union[:obj:`AttributeValueType`, :obj:`DataSourceType`]

        """
        arrays: list[Union[None, pa.Array]] = []
        size = None
        size_col = ""

        # We don't know the size of the table until we've found a
        # field in the schema which corresponds to a kwarg with data.
        # Therefore, we need to keep track of which columns are empty
        # *before* we've discovered the size of the table so we can
        # populate them with nulls later.
        empty_columns = []

        metadata: dict[bytes, bytes] = {}
        for i, field in enumerate(cls.schema):
            column_name = field.name
            column = getattr(cls, column_name)

            value = kwargs.pop(column_name, None)
            array = column._load(value, size)
            if array is None:
                # We'll have to wait until we get to a non-None column
                # to figure out the size.
                empty_columns.append(i)
                arrays.append(None)
                continue

            if size is None:
                size = len(array)
                size_col = column_name
            elif len(array) != size:
                raise errors.InvalidColumnDataError(
                    column,
                    f"wrong length {len(value)} (expected {size} based on column {size_col})",
                )

            arrays.append(array)

            if isinstance(value, Table):
                field_meta = value.table.schema.metadata
                if field_meta is not None:
                    for key, val in field_meta.items():
                        key = (field.name + "." + key.decode("utf-8")).encode("utf-8")
                        metadata[key] = val

        if size is None:
            raise ValueError("No data provided")

        for idx in empty_columns:
            column = getattr(cls, cls.schema[idx].name)
            arrays[idx] = column._nulls(size)

        # Inform the type checker that we've filled all Nones
        arrays = cast(list[pa.Array], arrays)

        for i, array in enumerate(arrays):
            if array.null_count > 0:
                column = getattr(cls, cls.schema[i].name)
                arrays[i] = column.fill_default(array)

        pyarrow_table = cls._build_arrow_table(arrays, metadata)
        attrib_kwargs = cls._attribute_kwargs_from_kwargs(kwargs)
        return cls.from_pyarrow(table=pyarrow_table, validate=validate, **attrib_kwargs)

    @classmethod
    def _build_arrow_table(cls, arrays: List[pa.Array], metadata: dict[bytes, bytes]) -> pa.Table:
        """
        Construct a pyarrow Table which will back cls from a list of arrays. The
        Table's schema comes from cls.schema.
        """
        schema = cls.schema.with_metadata(metadata)
        table = pa.Table.from_arrays(arrays, schema=schema)
        return table

    @classmethod
    def from_arrays(
        cls,
        arrays: list[pa.Array],
        metadata: Optional[dict[bytes, bytes]] = None,
        **kwargs: AttributeValueType,
    ) -> Self:
        """
        Create a Table object from a list of arrays.

        :param arrays: A list of pyarrow.Array objects.
        :param metadata: An optional dictionary of metadata to attach to the Table.
        :param \\**kwargs: Additional keyword arguments for any Table attributes.
        :type \\**kwargs: :obj:`AttributeValueType`
        :return: A Table object.
        """
        warnings.warn(DeprecationWarning("Table.from_arrays will be removed in quivr version 0.7"))
        if metadata is None:
            metadata = {}
        table = cls._build_arrow_table(arrays, metadata)
        return cls.from_pyarrow(table=table, validate=False, **kwargs)

    @classmethod
    def from_pydict(
        cls, d: dict[str, Union[pa.array, list[Any], npt.NDArray[Any]]], **kwargs: AttributeValueType
    ) -> Self:
        warnings.warn(DeprecationWarning("Table.from_pydict will be removed in quivr version 0.7"))
        table = pa.Table.from_pydict(d, schema=cls.schema)
        return cls.from_pyarrow(table=table, validate=False, **kwargs)

    @classmethod
    def from_rows(cls, rows: list[dict[str, Any]], **kwargs: AttributeValueType) -> Self:
        """
        Create a Table object from a list of dictionaries.

        :param rows: A list of values. Each value corresponds to a row in the table.
        :param \\**kwargs: Additional keyword arguments for any Table attributes.
        :type \\**kwargs: :obj:`AttributeValueType`
        :returns: A Table object.
        """
        warnings.warn(DeprecationWarning("Table.from_rows will be removed in quivr version 0.7"))
        table = pa.Table.from_pylist(rows, schema=cls.schema)
        return cls(table=table, **kwargs)

    @classmethod
    def from_lists(cls, lists: list[list[Any]], **kwargs: AttributeValueType) -> Self:
        """Create a Table object from a list of lists.

        Each inner list corresponds to a column in the Table. They
        should be specified in the same order as the columns in the
        class.

        :param lists: A list of lists. Each inner list corresponds to a column in the table.
        :param \\**kwargs: Additional keyword arguments for any Table attributes.
        :type \\**kwargs: :obj:`AttributeValueType`
        :returns: A Table object.

        """
        warnings.warn(DeprecationWarning("Table.from_lists will be removed in quivr version 0.7"))
        arrays = list(map(pa.array, lists))
        table = cls._build_arrow_table(arrays, {})
        return cls.from_pyarrow(table=table, validate=False, **kwargs)

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame, validate: bool = True, **kwargs: AttributeValueType) -> Self:
        """Load a DataFrame into the Table.

        If the DataFrame is missing any of the Table's columns, an
        error is raised. If the DataFrame has extra columns, they are
        ignored.

        This function cannot load "flattened" dataframes. This only
        matters for nested Tables which contain other Table
        definitions as columns. For that use case, either load an
        unflattened DataFrame, or use from_flat_dataframe.

        :param df: A pandas DataFrame containing the data to load.
        :param validate: Whether to validate the data after loading.
        :param \\**kwargs: Additional keyword arguments for any Table attributes.
        :type \\**kwargs: :obj:`AttributeValueType`
        """

        table = pa.Table.from_pandas(df, schema=cls.schema)
        return cls.from_pyarrow(table=table, validate=validate, **kwargs)

    @classmethod
    def _unflatten_table(cls, table: pa.Table) -> pa.Table:
        """Unflatten a Table.

        This is used when loading a flattened CSV into a nested
        Table. It takes a Table with a flat schema, and returns a
        Table with a nested schema.
        """
        struct_fields = []

        for field in cls.schema:
            if pa.types.is_struct(field.type):
                struct_fields.append(field)

        # Walk the schema, and build a StructArray for each embedded
        # type.

        def struct_array_for(field: pa.Field, ancestors: list[pa.Field]) -> pa.StructArray:
            prefix = ".".join([f.name for f in ancestors if f.name] + [field.name])

            child_arrays = []
            for subfield in field.type:
                if pa.types.is_struct(subfield.type):
                    child_arrays.append(struct_array_for(subfield, ancestors + [field]))
                else:
                    path = prefix + "." + subfield.name
                    child_arrays.append(table.column(path).combine_chunks())
            return pa.StructArray.from_arrays(child_arrays, fields=list(field.type))

        child_arrays = []
        for field in cls.schema:
            if pa.types.is_struct(field.type):
                child_arrays.append(struct_array_for(field, []))
            else:
                child_arrays.append(table.column(field.name).combine_chunks())

        return pa.Table.from_arrays(child_arrays, schema=cls.schema)

    @classmethod
    def from_flat_dataframe(
        cls, df: pd.DataFrame, validate: bool = True, **kwargs: AttributeValueType
    ) -> Self:
        """Load a flattened DataFrame into the Table.

        .. caution::
          Known bug: Doesn't correctly interpret fixed-length lists.

        :param df: A pandas DataFrame containing the data to load.
        :param validate: Whether to validate the data after loading.
        :param \\**kwargs: Additional keyword arguments for any Table attributes.
        :type \\**kwargs: :obj:`AttributeValueType`

        """
        struct_fields = []
        for field in cls.schema:
            if pa.types.is_struct(field.type):
                struct_fields.append(field)

        if len(struct_fields) == 0:
            table = pa.Table.from_pandas(df, schema=cls.schema)
            return cls.from_pyarrow(table=table, validate=validate, **kwargs)

        root = pa.field("", pa.struct(cls.schema))

        # Walk the schema, and build a StructArray for each embedded
        # type. These are stored in a dictionary, keyed by their
        # dot-separated path. For example, if the schema is:
        #
        #   pa.field("foo", pa.struct([
        #       pa.field("bar", pa.struct([
        #           pa.field("baz", pa.int64())
        #       ]))
        #   ]))
        #
        # Then the struct array for the inner struct will be stored in
        # the dictionary at the key "foo.bar".

        struct_arrays: dict[str, pa.StructArray] = {}

        def visitor(field: pa.Field, ancestors: list[pa.Field]) -> None:
            # Modify the dataframe in place, trimming out the columns we
            # have already processed in depth-first order.
            nonlocal df
            if len(ancestors) == 0:
                # Root - gets special behavior.
                df_key = ""
            else:
                lineage = [f.name for f in ancestors if f.name] + [field.name]
                df_key = ".".join(lineage)

            # Pull out just the columns relevant to this field
            field_columns = df.columns[df.columns.str.startswith(df_key)]
            field_df = df[field_columns]

            # Replace column names like "foo.bar.baz" with "baz", the
            # last component.
            if len(ancestors) == 0:
                # If we're at the root, use the original column names.
                names = field_df.columns
            else:
                names = field_df.columns.str.slice(len(df_key) + 1)
            field_df.columns = names

            # Build a StructArray of all of the subfields.
            arrays = []
            for subfield in field.type:
                sa_key = df_key + "." + subfield.name if df_key else subfield.name
                if sa_key in struct_arrays:
                    # We've already built this array, so just use it.
                    arrays.append(struct_arrays[sa_key])
                else:
                    arrays.append(field_df[subfield.name])
            sa = pa.StructArray.from_arrays(arrays, fields=list(field.type))
            struct_arrays[df_key] = sa

            # Clean the fields out
            df = df.drop(field_columns, axis="columns")

        schemagraph._walk_schema(root, visitor, None)
        # Now build a table back up. Grab the root-level struct array.
        sa = struct_arrays[""]

        table_arrays = []
        for subfield in cls.schema:
            # Pull out the fields of that root-level struct array.
            table_arrays.append(sa.field(subfield.name))

        table = pa.Table.from_arrays(table_arrays, schema=cls.schema)
        return cls.from_pyarrow(table, validate=validate, **kwargs)

    def flattened_table(self) -> pa.Table:
        """Completely flatten the Table's underlying Arrow table,
        taking into account any nested structure, and return the data
        table itself.
        """

        table = self.table
        while any(isinstance(field.type, pa.StructType) for field in table.schema):
            table = table.flatten()
        return table

    def select(self, column_name: str, value: Any) -> Self:
        """Select from the table by exact match, returning a new
        Table which only contains rows for which the value in
        column_name equals value.

        :param column_name: The name of the column to select on.
        :param value: The value to match.
        """
        table = self.table.filter(pc.field(column_name) == value)
        return self.__class__(table)

    def sort_by(self, by: Union[str, list[tuple[str, str]]]) -> Self:
        """Sorts the Table by the given column name (or multiple
        columns). This operation requires a copy, and returns a new
        Table using the copied data.

        by should be a column name to sort by, or a list of (column,
        order) tuples, where order can be "ascending" or "descending".

        :param by: The column name or list of (column, order) tuples to sort by.
        """
        table = self.table.sort_by(by)
        return self.__class__(table)

    def chunk_counts(self) -> dict[str, int]:
        """Returns the number of discrete memory chunks that make up
        each of the Table's underlying arrays. The keys of the
        resulting dictionary are the column names, and the values are
        the number of chunks for that column's data.
        """
        result = {}
        for i, field in enumerate(self.schema):
            result[field.name] = self.table.column(i).num_chunks
        return result

    def fragmented(self) -> bool:
        """Returns true if the Table has any fragmented arrays. If
        this is the case, performance might be improved by calling
        defragment on it.

        """
        return any(v > 1 for v in self.chunk_counts().values())

    def to_structarray(self) -> pa.StructArray:
        """Returns self as a StructArray.

        This only works if self is not fragmented. Call table =
        defragment(table) if table.fragmented() is True.

        :raises TableFragmentedError: if the table is fragmented.
        """
        if self.fragmented():
            raise errors.TableFragmentedError(
                "Tables cannot be converted to StructArrays while fragmented; call defragment(table) first."
            )
        arrays = [chunked_array.chunks[0] for chunked_array in self.table.columns]
        return pa.StructArray.from_arrays(arrays, fields=list(self.schema))

    def to_dataframe(self, flatten: bool = True) -> pd.DataFrame:
        """Returns self as a pandas DataFrame.

        If flatten is true, then any nested hierarchy is flattened: if
        the Table's schema contains a struct named "foo" with field
        "a", "b", and "c", then the resulting DataFrame will include
        columns "foo.a", "foo.b", "foo.c". This is done fully for any
        deeply nested structure, for example "foo.bar.baz.c".

        If flatten is false, then that struct will be in a single
        "foo" column, and the values will of the column will be
        dictionaries representing the struct values.

        :param flatten: Whether to flatten the table's structure.
        """
        table = self.table
        if flatten:
            table = self.flattened_table()
        df: pd.DataFrame = table.to_pandas()
        return df

    def column(self, column_name: str) -> pa.ChunkedArray:
        """
        Returns the column with the given name as a raw pyarrow ChunkedArray.

        :param column_name: The name of the column to return.
        """
        return self.table.column(column_name)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(size={len(self.table)})"

    def __len__(self) -> int:
        """
        Returns the number of rows in the Table.
        """
        return len(self.table)

    def with_table(self, table: pa.Table) -> Self:
        return self.__class__(table)

    def __getitem__(self, idx: Union[int, slice]) -> Self:
        """
        Returns a new Table containing the given row or rows.

        :param idx: The row index or slice to return.
        """
        if isinstance(idx, int):
            table = self.table[idx : idx + 1]
        else:
            table = self.table[idx]
        return self.with_table(table)

    def __iter__(self) -> Iterator[Self]:
        """
        Iterates over the rows of the Table, returning a new Table
        containing each row.
        """
        for i in range(len(self)):
            yield self[i : i + 1]

    def __eq__(self, other: Any) -> bool:
        """Returns true if the two Tables are equal. They are
        considered equal if they have the same data in their tables,
        and identical attributes and attribute values.

        :param other: The other Table to compare to.

        """
        if isinstance(other, Table):
            if not bool(self.table.equals(other.table, check_metadata=False)):
                return False
            if not self._attr_equal(other):
                return False
            return True
        if isinstance(other, pa.Table):
            return bool(self.table.equals(other, check_metadata=True))

        return False

    def _attr_equal(self, other: Self) -> bool:
        """
        Check if the attributes of two tables are equal. Recurs into subtables.
        """
        if not isinstance(other, self.__class__):
            return False

        if self.attributes() != other.attributes():
            return False

        for subtable_name in self._quivr_subtables.keys():
            self_subtable = getattr(self, subtable_name)
            other_subtable = getattr(other, subtable_name)
            if other_subtable is None:
                return False
            if not self_subtable._attr_equal(other_subtable):
                return False
        return True

    def take(self, row_indices: Union[list[int], pa.IntegerArray]) -> Self:
        """
        Return a new Table with only the rows at the given indices.

        :param row_indices: The indices of the rows to return.
        """
        return self.__class__(self.table.take(row_indices))

    def to_parquet(self, path: str, **kwargs: Any) -> None:
        """
        Write the table to a Parquet file.

        :param path: The path to write the Parquet file to.
        :param kwargs: Additional arguments to pass to pyarrow.parquet.write_table.
        """
        pyarrow.parquet.write_table(self.table, path, **kwargs)

    @classmethod
    def from_parquet(
        cls,
        path: str,
        memory_map: bool = False,
        pq_buffer_size: int = 0,
        filters: Optional[pc.Expression] = None,
        column_name_map: Optional[dict[str, str]] = None,
        validate: bool = True,
        **kwargs: AttributeValueType,
    ) -> Self:
        """Read a table from a Parquet file.

        :param path: The path to the Parquet file.
        :param memory_map: If True, memory-map the file, otherwise read it into memory.

        :param pq_buffer_size: If positive, perform read buffering
                when deserializing individual column
                chunks. Otherwise, IO calls are unbuffered.
        :param filters: An optional filter predicate to apply to the
                data. Rows which do not match the predicate will be
                removed from scanned data. For more information, see
                the PyArrow documentation on
                pyarrow.parquet.read_table and its filter parameter.
        :param column_name_map: An optional dictionary mapping column names in the Parquet file to
                column names in the resulting Table. This is useful if the Parquet file contains
                column names that are not valid Python identifiers, or if you want to rename
                columns for any other reason.
        :param validate: Whether to run column validation on the resulting Table.
        :param \\**kwargs: Additional keyword arguments to pass to Self's __init__ method.

        """
        table = cls._load_parquet_table(
            path=path,
            memory_map=memory_map,
            pq_buffer_size=pq_buffer_size,
            filters=filters,
            column_name_map=column_name_map,
        )
        return cls.from_pyarrow(table=table, validate=validate, **kwargs)

    @classmethod
    def _load_parquet_table(
        cls,
        path: str,
        memory_map: bool,
        pq_buffer_size: int,
        filters: Optional[pc.Expression],
        column_name_map: Optional[dict[str, str]],
    ) -> pa.Table:
        if column_name_map is not None:
            for value in column_name_map.values():
                if value not in cls.schema.names:
                    raise ValueError(
                        f"Column name {value} does not exist for {cls.__name__}, so cannot rename to it"
                    )
            inverted_map = {v: k for k, v in column_name_map.items()}
            column_names = [inverted_map.get(field.name, field.name) for field in cls.schema]
            schema = pa.schema(
                [pa.field(inverted_map.get(field.name, field.name), field.type) for field in cls.schema]
            )
        else:
            column_names = [field.name for field in cls.schema]
            schema = cls.schema

        table = pyarrow.parquet.read_table(
            source=path,
            columns=column_names,
            memory_map=memory_map,
            buffer_size=pq_buffer_size,
            filters=filters,
            schema=schema,
        )
        md = pyarrow.parquet.read_metadata(path, memory_map=memory_map)
        table = table.replace_schema_metadata(md.metadata)

        if column_name_map is not None:
            table = table.rename_columns([field.name for field in cls.schema])

        return table

    def to_feather(self, path: str, **kwargs: Any) -> None:
        """
        Write the table to a Feather file.

        :param path: The path to write the Feather file to.
        :param kwargs: Additional arguments to pass to pyarrow.feather.write_feather.
        """
        pyarrow.feather.write_feather(self.table, path, **kwargs)

    @classmethod
    def from_feather(cls, path: str, validate: bool = True, **kwargs: AttributeValueType) -> Self:
        """Read a table from a Feather file.

        :param path: The path to the Feather file.
        :param validate: Whether to run column validators on the table after loading it.
        :param \\**kwargs: Additional keyword arguments to pass to Self's __init__ method.
        """
        table = pyarrow.feather.read_table(path)
        return cls.from_pyarrow(table=table, validate=validate, **kwargs)

    def to_csv(self, path: str, attribute_columns: bool = True) -> None:
        """Write the table to a CSV file. Any nested structure is flattened.

        :param path: The path to write the CSV file to.
        :param attribute_columns: If True, store any Attributes defined for the table
            (or its subtable columns) as columns in the CSV file. If False, do not store
            any Attribute data in the CSV file.
        """
        table = self.flattened_table()
        if attribute_columns:
            for name, val in self._string_attributes().items():
                table = table.append_column(
                    name,
                    pa.repeat(val, len(table)),
                )
        pyarrow.csv.write_csv(table, path)

    @classmethod
    def from_csv(
        cls,
        input_file: Union[str, os.PathLike, IOBase],  # type: ignore
        validate: bool = True,
        **kwargs: AttributeValueType,
    ) -> Self:
        """
        Read a table from a CSV file.

        :param input_file: The path to the CSV file, or a file-like object.
        :param \\**kwargs: Additional keyword arguments to set the Table's attributes.
        """
        flat_table = pyarrow.csv.read_csv(input_file)

        # Gather any attributes from the CSV. We do this by looking for specially named
        # columns, and grabbing the value from their first row.
        attributes = {}
        attr_names = cls._attribute_metadata_keys()
        if len(flat_table) > 0:
            to_be_removed = []
            for name in flat_table.column_names:
                if name in attr_names:
                    attributes[name] = str(flat_table.column(name)[0].as_py())
                    to_be_removed.append(name)

            if len(to_be_removed) > 0:
                flat_table = flat_table.drop_columns(to_be_removed)

        attribute_meta = cls._unpack_string_metadata(attributes)

        table = cls._unflatten_table(flat_table)
        metadata = table.schema.metadata or {}
        metadata.update(attribute_meta)
        table = table.replace_schema_metadata(metadata)
        return cls.from_pyarrow(table=table, validate=validate, **kwargs)

    def is_valid(self) -> bool:
        """Validate the table against the schema."""
        for name, validator in self._column_validators.items():
            if not validator.valid(self.table.column(name)):
                return False
        return True

    def validate(self) -> None:
        """Validate the table against the schema, raising an exception if invalid."""
        for name, validator in self._column_validators.items():
            try:
                validator.validate(self.table.column(name))
            except errors.ValidationError as e:
                raise errors.ValidationError(f"Column {name} failed validation: {str(e)}", e.failures) from e

    @classmethod
    def empty(cls, **kwargs: AttributeValueType) -> Self:
        """Create an empty instance of the table.

        :param \\**kwargs: Additional keyword arguments to set the Table's attributes.
        """
        data = [[] for _ in range(len(cls.schema))]  # type: ignore
        empty_table = pa.table(data, schema=cls.schema)
        return cls.from_pyarrow(table=empty_table, validate=False, **kwargs)

    def attributes(self) -> dict[str, Any]:
        """Return a dictionary of the table's attributes."""
        return {name: getattr(self, name) for name in self._quivr_attributes}

    def _string_attributes(self) -> dict[str, str]:
        """Return a dictionary of the table's attributes.

        Attributes are presented in their string form.
        """
        d = {}
        for name, descriptor in self._quivr_attributes.items():
            value = getattr(self, name)
            d[name] = descriptor.to_string(value)
        for name in self._quivr_subtables.keys():
            subattribs = getattr(self, name)._string_attributes()
            for subname, subval in subattribs.items():
                d[f"{name}.{subname}"] = subval
        return d

    @classmethod
    def _unpack_string_metadata(cls, metadata: dict[str, str]) -> dict[bytes, bytes]:
        result = {}

        for k, v in metadata.items():
            if "." in k:
                # This is metadata for a subtable; we need to dig into it to find the
                # descriptor for the attribute.
                subtable_name, subtable_key = k.split(".", 1)
                subtable_item = cls._quivr_subtables[subtable_name].table_type._unpack_string_metadata(
                    {subtable_key: v}
                )
                result[k.encode("utf8")] = next(iter(subtable_item.values()))
            else:
                descriptor = cls._quivr_attributes.get(k)
                if descriptor is None:
                    raise AttributeError(f"unexpected missing descriptor with name={k}")
                result[k.encode("utf8")] = descriptor.to_bytes(descriptor.from_string(v))
        return result

    def _metadata_for_column(self, column_name: str) -> dict[bytes, bytes]:
        """Return a dictionary of metadata associated with a subtable column."""
        result: dict[bytes, bytes] = {}
        if self.table.schema.metadata is None:
            return result
        column_name_bytes = (column_name + ".").encode("utf8")
        for key, value in self.table.schema.metadata.items():
            if key.startswith(column_name_bytes):
                result[key[len(column_name_bytes) :]] = value
        return result

    @classmethod
    def _attribute_metadata_keys(cls) -> set[str]:
        """Return a set of all subtable attribute names."""
        result = {attr.name for attr in cls._quivr_attributes.values()}
        for column in cls._quivr_subtables.values():
            attr_fields = column.table_type._quivr_attributes
            for attr in attr_fields.values():
                result.add(f"{column.name}.{attr.name}")

            children = column.table_type._attribute_metadata_keys()
            for key in children:
                result.add(f"{column.name}.{key}")
        return result

    def apply_mask(self, mask: pa.BooleanArray | np.ndarray[bool, Any] | list[bool]) -> Self:
        """
        Return a new table with rows filtered to match a boolean mask.

        The mask must have the same length as the table. At each index, if the mask's
        value is True, the row will be included in the new table; if False, it will be
        excluded.

        If the mask is a pyarrow BooleanArray, it must not have any null values.

        :param mask: A boolean mask to apply to the table.
        """
        if len(mask) != len(self):
            raise ValueError("mask must be the same length as the table")
        if isinstance(mask, pa.BooleanArray):
            if mask.null_count > 0:
                raise ValueError("mask must not contain null values")
        return self.__class__(self.table.filter(mask))

    def where(self, expr: pc.Expression) -> Self:
        """
        Return a new table with rows filtered to match an expression.

        The expression must be a pyarrow Expression that evaluates to a boolean array.

        :param expr: A pyarrow Expression to apply to the table.

        Examples:
            >>> import quivr as qv
            >>> import pyarrow.compute as pc
            >>> class MyTable(qv.Table):
            ...     x = qv.Int64Column()
            ...     y = qv.Int64Column()
            >>> t = MyTable.from_kwargs(x=[1, 2, 3], y=[4, 5, 6])
            >>> filtered = t.where(pc.field("x") > 1)
            >>> print(filtered.x.to_pylist())
            [2, 3]
            >>> print(filtered.y.to_pylist())
            [5, 6]
        """
        return self.__class__(self.table.filter(expr))

    def __arrow_array__(self, type: Optional[pa.DataType] = None) -> pa.StructArray:
        """
        Implements the Arrow array protocol by returning as StructArray.
        """
        array = self.to_structarray()
        if type is None:
            return array
        return array.cast(type)
